import numpy as np
import pyopencl as cl
from PIL import Image
from .kernel import kernel
from ..shared import Bounds, ImageConfig, GPU


def _init_arrays(step_size: float, bounds: Bounds) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    real_vals = np.arange(bounds.x_min, bounds.x_max, step_size, dtype=np.float32)
    imag_vals = np.arange(bounds.y_min, bounds.y_max, step_size, dtype=np.float32)
    c_real, c_imag = np.meshgrid(real_vals, imag_vals)
    c_real, c_imag = c_real.flatten(), c_imag.flatten()
    z_real = np.zeros(len(c_real), dtype=np.uint32)
    z_imag = np.zeros(len(c_real), dtype=np.uint32)
    return c_real, c_imag, z_real, z_imag


def _get_array_buffer(gpu: GPU, arr: np.ndarray, read_only: bool = False) -> cl.Buffer:
    if read_only:
        return cl.Buffer(gpu.ctx, gpu.mf.READ_ONLY | gpu.mf.COPY_HOST_PTR, hostbuf=arr)
    return cl.Buffer(gpu.ctx, gpu.mf.READ_WRITE | gpu.mf.COPY_HOST_PTR, hostbuf=arr)


def _collect_array(gpu: GPU, buffer: cl.Buffer, arr: np.ndarray):
    cl.enqueue_copy(gpu.queue, arr, buffer).wait()


def nebulabrot(
        gpu: GPU,
        image_config: ImageConfig,
        equation: str,
        step_size: float,
        max_iter: int,
        bounds: Bounds,
        bail_mag: float = 4.0
    ) -> Image:
    image_data = np.zeros(image_config.height * image_config.width, dtype=np.uint32)
    c_real, c_imag, z_real, z_imag = _init_arrays(step_size, bounds)
    iter_count = 0
    while iter_count < max_iter:
        current_iter = min(10000, max_iter - iter_count)
        d_c_real = _get_array_buffer(gpu, c_real, read_only=True)
        d_c_imag = _get_array_buffer(gpu, c_imag, read_only=True)
        d_z_real = _get_array_buffer(gpu, z_real)
        d_z_imag = _get_array_buffer(gpu, z_imag)
        d_image_data = _get_array_buffer(gpu, image_data)
        kernel_str = kernel(image_config, equation, current_iter, bail_mag, bounds)
        program = cl.Program(gpu.ctx, kernel_str).build()
        program.render(gpu.queue, (len(c_real),), None, d_c_real, d_c_imag, d_z_real, d_z_imag, d_image_data)
        _collect_array(gpu, d_z_real, z_real)
        _collect_array(gpu, d_z_imag, z_imag)
        _collect_array(gpu, d_image_data, image_data)

        valid = z_real**2 + z_imag**2 < bail_mag
        c_real = c_real[valid]
        c_imag = c_imag[valid]
        z_real = z_real[valid]
        z_imag = z_imag[valid]

        if len(c_real) == 0:
            break

        iter_count += current_iter

    image_data = image_config.dr_func(image_data)
    image_data = ((image_data / np.max(image_data)) * image_config.max_val).astype(image_config.numpy_dtype)
    image_data = image_data.reshape(image_config.width, image_config.height)
    return Image.fromarray(image_data, mode=image_config.pil_mode)


from ..shared import Bounds, ImageConfig


def kernel(
        image_config: ImageConfig,
        equation: str,
        max_iter: int,
        bail_mag: float,
        bounds: Bounds
    ) -> str:
    return """
typedef struct {{
    float real;
    float imag;
}} Complex;

Complex add(Complex a, Complex b) {{
    Complex result;
    result.real = a.real + b.real;
    result.imag = a.imag + b.imag;
    return result;
}}

Complex subtract(Complex a, Complex b) {{
    Complex result;
    result.real = a.real - b.real;
    result.imag = a.imag - b.imag;
    return result;
}}

Complex multiply(Complex a, Complex b) {{
    Complex result;
    result.real = a.real * b.real - a.imag * b.imag;
    result.imag = a.real * b.imag + a.imag * b.real;
    return result;
}}

Complex divide(Complex a, Complex b) {{
    float denom = b.real * b.real + b.imag * b.imag;
    Complex result;
    result.real = (a.real * b.real + a.imag * b.imag) / denom;
    result.imag = (a.imag * b.real - a.real * b.imag) / denom;
    return result;
}}

float mag_squared(Complex z) {{
    return z.real * z.real + z.imag * z.imag;
}}


__kernel void render(
    __global float *c_real, __global float *c_imag,
    __global float *z_real, __global float *z_imag,
    __global int *image_data
) 
{{
    int i = get_global_id(0);
    Complex c = {{ c_real[i], c_imag[i] }};
    Complex z = {{ z_real[i], z_imag[i] }};
    Complex locations[{max_iter}];
    int count = 0;

    for (int iter = 0; iter < {max_iter}; iter++) {{
        z = {equation};
        locations[iter] = z;
        count++;
        if (mag_squared(z) > {bail_mag}f) break;
    }}

    if (mag_squared(z) > {bail_mag}f) {{
        for (int j = 0; j < count; j++) {{
            int px = (int)((locations[j].real - {x_min}) / ({x_max} - {x_min}) * {width});
            int py = (int)((locations[j].imag - {y_min}) / ({y_max} - {y_min}) * {height});

            if (px >= 0 && px < {width} && py >= 0 && py < {height}) {{
                atomic_inc(&image_data[py * {width} + px]);
            }}
        }}
    }} else {{
        z_real[i] = z.real;
        z_imag[i] = z.imag;
    }}
}}
""".format(
    equation=equation,
    max_iter=max_iter,
    bail_mag=bail_mag,
    x_min=bounds.x_min,
    x_max=bounds.x_max,
    y_min=bounds.y_min,
    y_max=bounds.y_max,
    width=image_config.width,
    height=image_config.height
)


from enum import Enum
from typing import Callable
import numpy as np
import pyopencl as cl


class GPU:
    def __init__(self):
        self.ctx = cl.create_some_context()
        self.queue = cl.CommandQueue(self.ctx)
        self.mf = cl.mem_flags


class Bounds:

    def __init__(self, x_min: float, x_max: float, y_min: float, y_max: float):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max


class BitDepth(Enum):
    EIGHT = 8
    SIXTEEN = 16
    THIRTY_TWO = 32


class DynamicRangeBoost(Enum):
    log = "log"
    sqrt = "sqrt"


class ImageConfig:

    def __init__(self, width: int, height: int, bit_depth: BitDepth, dynamic_range_boost: DynamicRangeBoost):
        self.width = width
        self.height = height
        self.bit_depth = bit_depth
        self.dynamic_range_boost = dynamic_range_boost

    @property
    def max_val(self) -> int:
        return 2 ** self.bit_depth.value - 1
    
    @property
    def numpy_dtype(self):
        numpy_dtype_mapping = {
            BitDepth.EIGHT: np.uint8,
            BitDepth.SIXTEEN: np.uint16,
            BitDepth.THIRTY_TWO: np.uint32
        }
        return numpy_dtype_mapping[self.bit_depth]
    
    @property
    def pil_mode(self) -> str:
        pil_mode_mapping = {
            BitDepth.EIGHT: 'L',
            BitDepth.SIXTEEN: 'I;16',
            BitDepth.THIRTY_TWO: 'I'
        }
        return pil_mode_mapping[self.bit_depth]
    
    @property
    def dr_func(self) -> Callable:
        dr_func_mapping = {
            DynamicRangeBoost.log: np.log1p,
            DynamicRangeBoost.sqrt: np.sqrt
        }
        return dr_func_mapping[self.dynamic_range_boost]





from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read().splitlines()

setup(
    name = 'iterart',
    version = '0.1.0',
    author = "Carson O'Ffill",
    author_email = 'offillcarson@gmail.com',
    license = 'MIT',
    description = 'A collection of tools to generate renderings based on iterative functions.',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/karzunn/iterart',
    py_modules = ['iterart'],
    packages = find_packages(),
    install_requires = [requirements],
    python_requires='>=3.12',
    classifiers=[
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ]
)